"""
A datalog reader.
"""

from collections import defaultdict
from itertools import chain

from datalog.parser import FAILURE, Grammar
from datalog.types import Constant, Dataset, LVar, Rule


class Actions(object):
  def __init__(self, db_cls=None):
    self._db_cls = db_cls or Dataset

  def make_dataset(self, input, start, end, elements):
    # Group the various terms
    rules = []
    tuples = []
    for e in elements:
      if e:
        if isinstance(e, Rule):
          rules.append(e)
        else:
          tuples.append(e)

    return self._db_cls(tuples, rules)

  def make_symbol(self, input, start, end, elements):
    return LVar("".join(e.text for e in elements),)

  def make_word(self, input, start, end, elements):
    return Constant("".join(e.text for e in elements),)

  def make_string(self, input, start, end, elements):
    return Constant(elements[1].text,)

  def make_comment(self, input, start, end, elements):
    return None

  def make_ws(self, input, start, end, elements=None):  # watf?
    pass

  def make_rule(self, input, start, end, elements):
    if elements[1].elements:
      return Rule(elements[0], elements[1].elements[3][1])
    else:
      return elements[0]

  def make_clause(self, input, start, end, elements):
    if elements[0].text == "~":
      return ("not", (elements[1], *elements[3][1]))
    else:
      return (elements[1], *elements[3][1])

  def make_terms(self, input, start, end, elements):
    return self._make("terms", elements)

  def make_clauses(self, input, start, end, elements):
    return self._make("clauses", elements)

  def _make(self, tag, elements):
    if len(elements) == 1:
      return (tag, [elements[0]],)
    elif elements[1].elements:
      return (tag, [elements[0]] + elements[1].elements[2][1])
    else:
      return (tag, [elements[0]])

  def make_command(self, input, start, end, elements):
    op = elements[-1].text
    val = self.make_rule(input, start, end, elements)
    if op == ".":
      val = self.make_dataset(input, start, end, [val])
    return op, val


class Parser(Grammar):
  """Implementation detail.

  A slightly hacked version of the Parser class canopy generates, which lets us control what the
  parsing entry point is. This lets me play games with having one parser and one grammar which is
  used both for the command shell and for other things.

  """
  def __init__(self, input, actions, types):
    self._input = input
    self._input_size = len(input)
    self._actions = actions
    self._types = types
    self._offset = 0
    self._cache = defaultdict(dict)
    self._failure = 0
    self._expected = []

  def parse(self, meth):
    tree = meth()
    if tree is not FAILURE and self._offset == self._input_size:
      return tree
    if not self._expected:
      self._failure = self._offset
      self._expected.append("<EOF>")
      raise ParseError(format_error(self._input, self._failure, self._expected))


def format_error(input, offset, expected):
  lines, line_no, position = input.split("\n"), 0, 0
  while position <= offset:
    position += len(lines[line_no]) + 1
    line_no += 1
    message, line = "Line " + str(line_no) + ": expected " + ", ".join(expected) + "\n", lines[line_no - 1]
    message += line + "\n"
    position -= len(line) + 1
    message += " " * (offset - position)
  return message + "^"


def read_dataset(text: str, db_cls=None):
  """Read a string of text, returning a whole Datalog dataset."""

  parser = Parser(text, Actions(db_cls=db_cls), None)
  return parser.parse(parser._read_dataset)


def read_command(text: str, db_cls=None):
  """Read a string of text, returning a whole Datalog dataset."""

  actions = Actions(db_cls=db_cls)
  parser = Parser(text, actions, None)
  return parser.parse(parser._read_command)


read = read_dataset



from collections import defaultdict
import re


class TreeNode(object):
    def __init__(self, text, offset, elements=None):
        self.text = text
        self.offset = offset
        self.elements = elements or []

    def __iter__(self):
        for el in self.elements:
            yield el


class TreeNode1(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode1, self).__init__(text, offset, elements)
        self.clause = elements[0]


class TreeNode2(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode2, self).__init__(text, offset, elements)
        self.ws = elements[2]
        self.clauses = elements[3]


class TreeNode3(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode3, self).__init__(text, offset, elements)
        self.clause = elements[0]


class TreeNode4(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode4, self).__init__(text, offset, elements)
        self.ws = elements[1]
        self.clauses = elements[2]


class TreeNode5(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode5, self).__init__(text, offset, elements)
        self.word = elements[1]
        self.terms = elements[3]


class TreeNode6(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode6, self).__init__(text, offset, elements)
        self.term = elements[0]


class TreeNode7(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode7, self).__init__(text, offset, elements)
        self.ws = elements[1]
        self.terms = elements[2]


class TreeNode8(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode8, self).__init__(text, offset, elements)
        self.clause = elements[0]


class TreeNode9(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode9, self).__init__(text, offset, elements)
        self.ws = elements[2]
        self.clauses = elements[3]


class ParseError(SyntaxError):
    pass


FAILURE = object()


class Grammar(object):
    REGEX_1 = re.compile('^[A-Z]')
    REGEX_2 = re.compile('^[a-z0-9-_=<>]')
    REGEX_3 = re.compile('^[a-z0-9-_=<>]')
    REGEX_4 = re.compile('^[^\\"]')
    REGEX_5 = re.compile('^[^\\n]')
    REGEX_6 = re.compile('^[ \\t\\n]')

    def _read_dataset(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['dataset'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        remaining0, index1, elements0, address1 = 0, self._offset, [], True
        while address1 is not FAILURE:
            index2 = self._offset
            address1 = self._read_rule()
            if address1 is FAILURE:
                self._offset = index2
                address1 = self._read_comment()
                if address1 is FAILURE:
                    self._offset = index2
                    address1 = self._read_whitespace()
                    if address1 is FAILURE:
                        self._offset = index2
            if address1 is not FAILURE:
                elements0.append(address1)
                remaining0 -= 1
        if remaining0 <= 0:
            address0 = self._actions.make_dataset(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        else:
            address0 = FAILURE
        self._cache['dataset'][index0] = (address0, self._offset)
        return address0

    def _read_rule(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['rule'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_clause()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            index3, elements1 = self._offset, []
            address3 = FAILURE
            address3 = self._read_ws()
            if address3 is not FAILURE:
                elements1.append(address3)
                address4 = FAILURE
                chunk0 = None
                if self._offset < self._input_size:
                    chunk0 = self._input[self._offset:self._offset + 2]
                if chunk0 == ':-':
                    address4 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                    self._offset = self._offset + 2
                else:
                    address4 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('":-"')
                if address4 is not FAILURE:
                    elements1.append(address4)
                    address5 = FAILURE
                    address5 = self._read_ws()
                    if address5 is not FAILURE:
                        elements1.append(address5)
                        address6 = FAILURE
                        address6 = self._read_clauses()
                        if address6 is not FAILURE:
                            elements1.append(address6)
                        else:
                            elements1 = None
                            self._offset = index3
                    else:
                        elements1 = None
                        self._offset = index3
                else:
                    elements1 = None
                    self._offset = index3
            else:
                elements1 = None
                self._offset = index3
            if elements1 is None:
                address2 = FAILURE
            else:
                address2 = TreeNode2(self._input[index3:self._offset], index3, elements1)
                self._offset = self._offset
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address7 = FAILURE
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 == '.':
                    address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address7 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"."')
                if address7 is not FAILURE:
                    elements0.append(address7)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_rule(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['rule'][index0] = (address0, self._offset)
        return address0

    def _read_clauses(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['clauses'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_clause()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            index3, elements1 = self._offset, []
            address3 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == ',':
                address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address3 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('","')
            if address3 is not FAILURE:
                elements1.append(address3)
                address4 = FAILURE
                address4 = self._read_ws()
                if address4 is not FAILURE:
                    elements1.append(address4)
                    address5 = FAILURE
                    address5 = self._read_clauses()
                    if address5 is not FAILURE:
                        elements1.append(address5)
                    else:
                        elements1 = None
                        self._offset = index3
                else:
                    elements1 = None
                    self._offset = index3
            else:
                elements1 = None
                self._offset = index3
            if elements1 is None:
                address2 = FAILURE
            else:
                address2 = TreeNode4(self._input[index3:self._offset], index3, elements1)
                self._offset = self._offset
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_clauses(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['clauses'][index0] = (address0, self._offset)
        return address0

    def _read_clause(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['clause'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        index2 = self._offset
        address1 = self._read_negation()
        if address1 is FAILURE:
            address1 = TreeNode(self._input[index2:index2], index2)
            self._offset = index2
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            address2 = self._read_word()
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                chunk0 = None
                if self._offset < self._input_size:
                    chunk0 = self._input[self._offset:self._offset + 1]
                if chunk0 == '(':
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"("')
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address4 = FAILURE
                    address4 = self._read_terms()
                    if address4 is not FAILURE:
                        elements0.append(address4)
                        address5 = FAILURE
                        chunk1 = None
                        if self._offset < self._input_size:
                            chunk1 = self._input[self._offset:self._offset + 1]
                        if chunk1 == ')':
                            address5 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                            self._offset = self._offset + 1
                        else:
                            address5 = FAILURE
                            if self._offset > self._failure:
                                self._failure = self._offset
                                self._expected = []
                            if self._offset == self._failure:
                                self._expected.append('")"')
                        if address5 is not FAILURE:
                            elements0.append(address5)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_clause(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['clause'][index0] = (address0, self._offset)
        return address0

    def _read_negation(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['negation'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '~':
            address0 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address0 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"~"')
        self._cache['negation'][index0] = (address0, self._offset)
        return address0

    def _read_terms(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['terms'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_term()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            index3, elements1 = self._offset, []
            address3 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == ',':
                address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address3 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('","')
            if address3 is not FAILURE:
                elements1.append(address3)
                address4 = FAILURE
                address4 = self._read_ws()
                if address4 is not FAILURE:
                    elements1.append(address4)
                    address5 = FAILURE
                    address5 = self._read_terms()
                    if address5 is not FAILURE:
                        elements1.append(address5)
                    else:
                        elements1 = None
                        self._offset = index3
                else:
                    elements1 = None
                    self._offset = index3
            else:
                elements1 = None
                self._offset = index3
            if elements1 is None:
                address2 = FAILURE
            else:
                address2 = TreeNode7(self._input[index3:self._offset], index3, elements1)
                self._offset = self._offset
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_terms(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['terms'][index0] = (address0, self._offset)
        return address0

    def _read_term(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['term'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1 = self._offset
        address0 = self._read_string()
        if address0 is FAILURE:
            self._offset = index1
            address0 = self._read_lvar()
            if address0 is FAILURE:
                self._offset = index1
                address0 = self._read_word()
                if address0 is FAILURE:
                    self._offset = index1
        self._cache['term'][index0] = (address0, self._offset)
        return address0

    def _read_lvar(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['lvar'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 is not None and Grammar.REGEX_1.search(chunk0):
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('[A-Z]')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 is not None and Grammar.REGEX_2.search(chunk1):
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('[a-z0-9-_=<>]')
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_symbol(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['lvar'][index0] = (address0, self._offset)
        return address0

    def _read_word(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['word'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        remaining0, index1, elements0, address1 = 0, self._offset, [], True
        while address1 is not FAILURE:
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 is not None and Grammar.REGEX_3.search(chunk0):
                address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address1 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('[a-z0-9-_=<>]')
            if address1 is not FAILURE:
                elements0.append(address1)
                remaining0 -= 1
        if remaining0 <= 0:
            address0 = self._actions.make_word(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        else:
            address0 = FAILURE
        self._cache['word'][index0] = (address0, self._offset)
        return address0

    def _read_string(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['string'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '"':
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"\\""')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 is not None and Grammar.REGEX_4.search(chunk1):
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('[^\\"]')
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
                address4 = FAILURE
                chunk2 = None
                if self._offset < self._input_size:
                    chunk2 = self._input[self._offset:self._offset + 1]
                if chunk2 == '"':
                    address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address4 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"\\""')
                if address4 is not FAILURE:
                    elements0.append(address4)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_string(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['string'][index0] = (address0, self._offset)
        return address0

    def _read_ws(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['ws'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        remaining0, index1, elements0, address1 = 1, self._offset, [], True
        while address1 is not FAILURE:
            index2 = self._offset
            address1 = self._read_comment()
            if address1 is FAILURE:
                self._offset = index2
                address1 = self._read_whitespace()
                if address1 is FAILURE:
                    self._offset = index2
            if address1 is not FAILURE:
                elements0.append(address1)
                remaining0 -= 1
        if remaining0 <= 0:
            address0 = TreeNode(self._input[index1:self._offset], index1, elements0)
            self._offset = self._offset
        else:
            address0 = FAILURE
        self._cache['ws'][index0] = (address0, self._offset)
        return address0

    def _read_comment(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['comment'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '%':
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"%"')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 is not None and Grammar.REGEX_5.search(chunk1):
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('[^\\n]')
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
                address4 = FAILURE
                chunk2 = None
                if self._offset < self._input_size:
                    chunk2 = self._input[self._offset:self._offset + 1]
                if chunk2 == '\n':
                    address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address4 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"\\n"')
                if address4 is not FAILURE:
                    elements0.append(address4)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_comment(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['comment'][index0] = (address0, self._offset)
        return address0

    def _read_whitespace(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['whitespace'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        remaining0, index1, elements0, address1 = 1, self._offset, [], True
        while address1 is not FAILURE:
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 is not None and Grammar.REGEX_6.search(chunk0):
                address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address1 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('[ \\t\\n]')
            if address1 is not FAILURE:
                elements0.append(address1)
                remaining0 -= 1
        if remaining0 <= 0:
            address0 = self._actions.make_ws(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        else:
            address0 = FAILURE
        self._cache['whitespace'][index0] = (address0, self._offset)
        return address0

    def _read_command(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['command'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_clause()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            index3, elements1 = self._offset, []
            address3 = FAILURE
            address3 = self._read_ws()
            if address3 is not FAILURE:
                elements1.append(address3)
                address4 = FAILURE
                chunk0 = None
                if self._offset < self._input_size:
                    chunk0 = self._input[self._offset:self._offset + 2]
                if chunk0 == ':-':
                    address4 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                    self._offset = self._offset + 2
                else:
                    address4 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('":-"')
                if address4 is not FAILURE:
                    elements1.append(address4)
                    address5 = FAILURE
                    address5 = self._read_ws()
                    if address5 is not FAILURE:
                        elements1.append(address5)
                        address6 = FAILURE
                        address6 = self._read_clauses()
                        if address6 is not FAILURE:
                            elements1.append(address6)
                        else:
                            elements1 = None
                            self._offset = index3
                    else:
                        elements1 = None
                        self._offset = index3
                else:
                    elements1 = None
                    self._offset = index3
            else:
                elements1 = None
                self._offset = index3
            if elements1 is None:
                address2 = FAILURE
            else:
                address2 = TreeNode9(self._input[index3:self._offset], index3, elements1)
                self._offset = self._offset
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address7 = FAILURE
                index4 = self._offset
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 == '.':
                    address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address7 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"."')
                if address7 is FAILURE:
                    self._offset = index4
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 == '?':
                        address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address7 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"?"')
                    if address7 is FAILURE:
                        self._offset = index4
                        chunk3 = None
                        if self._offset < self._input_size:
                            chunk3 = self._input[self._offset:self._offset + 1]
                        if chunk3 == '!':
                            address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                            self._offset = self._offset + 1
                        else:
                            address7 = FAILURE
                            if self._offset > self._failure:
                                self._failure = self._offset
                                self._expected = []
                            if self._offset == self._failure:
                                self._expected.append('"!"')
                        if address7 is FAILURE:
                            self._offset = index4
                if address7 is not FAILURE:
                    elements0.append(address7)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_command(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['command'][index0] = (address0, self._offset)
        return address0


class Parser(Grammar):
    def __init__(self, input, actions, types):
        self._input = input
        self._input_size = len(input)
        self._actions = actions
        self._types = types
        self._offset = 0
        self._cache = defaultdict(dict)
        self._failure = 0
        self._expected = []

    def parse(self):
        tree = self._read_dataset()
        if tree is not FAILURE and self._offset == self._input_size:
            return tree
        if not self._expected:
            self._failure = self._offset
            self._expected.append('<EOF>')
        raise ParseError(format_error(self._input, self._failure, self._expected))


def format_error(input, offset, expected):
    lines, line_no, position = input.split('\n'), 0, 0
    while position <= offset:
        position += len(lines[line_no]) + 1
        line_no += 1
    message, line = 'Line ' + str(line_no) + ': expected ' + ', '.join(expected) + '\n', lines[line_no - 1]
    message += line + '\n'
    position -= len(line) + 1
    message += ' ' * (offset - position)
    return message + '^'

def parse(input, actions=None, types=None):
    parser = Parser(input, actions, types)
    return parser.parse()



"""
The core IR types for datalog.
"""

from collections import namedtuple
from typing import Sequence, Tuple, Union


class Constant(namedtuple("Constant", ["value"])):
  """Representation of a constant for interpreter dispatching."""


class LVar(namedtuple("LVar", ["name"])):
  """Representation of an LVar for interpreter dispatching."""


class Rule(namedtuple("Rule", ["pattern", "clauses"])):
  """Representation of an Rule for the interpreter."""

  def __new__(cls, pattern, clauses):
    return super(cls, Rule).__new__(cls, pattern, tuple(clauses))


# Logical type for 'Tuple' as we're gonna use it for the rest of the thing.
LTuple = Tuple[Union[Constant, LVar]]
CTuple = Tuple[Constant]


class Dataset(object):
  """A set of tuples and rules which can be queried."""

  def __init__(self,
               tuples: Sequence[CTuple],
               rules: Sequence[Rule]):
    self.__tuples = tuples
    self.__rules = rules

  def tuples(self) -> Sequence[CTuple]:
    for t in self.__tuples:
      yield t

  def rules(self) -> Sequence[Rule]:
    for r in self.__rules:
      yield r

  def merge(self, other: "Dataset") -> "Dataset":
    """Merge two datasets together, returning a new one."""

    return type(self)(list({*self.tuples(), *other.tuples()}),
                      [*self.rules(), *other.rules()])


class CachedDataset(Dataset):
  """An extension of the dataset which features a cache of rule produced tuples.

  Note that this cache is lost when merging datasets - which ensures correctness.
  """

  def __init__(self, tuples, rules):
    super(CachedDataset, self).__init__(tuples, rules)
    # The cache is a mapping from a Rule to tuples produced by it.
    self.__cache = {}

  def tuples(self) -> Sequence[CTuple]:
    yield from super(CachedDataset, self).tuples()

  def scan_cache(self, rule_tuple):
    if rule_tuple in self.__cache:
      return iter(self.__cache.get(rule_tuple))

  def cache_tuple(self, rule_tuple, tuple: CTuple):
    coll = self.__cache.get(rule_tuple, set())
    self.__cache[rule_tuple] = coll
    coll.add(tuple)


class IndexedDataset(CachedDataset):
  """An extension of the Dataset type which features both a cache and an index.

  The index allows more efficient scans by maintaining 'table' style partitions.
  It does not (yet) support user-defined indexing schemes.
  """

  @staticmethod
  def __key(t: LTuple) -> str:
    assert isinstance(t[0], Constant)
    return f"{t[0].value}_{len(t)}"

  def __init__(self, tuples, rules):
    super(IndexedDataset, self).__init__(tuples, rules)

    index = {}
    for t in self.tuples():
      key = self.__key(t)
      coll = index.get(key, list())
      # FIXME: Walrus operator???
      index[key] = coll
      coll.append(t)

    self.__index = index

  def scan_index(self, t: LTuple) -> Sequence[CTuple]:
    for t in self.__index.get(self.__key(t), []):
      yield t



"""Query evaluation unit tests."""

import pytest

from datalog.easy import read, select
from datalog.types import (CachedDataset, Constant, Dataset, IndexedDataset,
                           LVar, Rule)

DBCLS = [Dataset, CachedDataset, IndexedDataset]

@pytest.mark.parametrize("db_cls,", DBCLS)
def test_id_query(db_cls):
  """Querying for a constant in the dataset."""

  ab = (Constant("a"), Constant("b"),)
  assert not select(db_cls([], []), ("a", "b",))
  assert select(db_cls([ab], []), ("a", "b",)) == \
    [(("a", "b"), {},)]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_lvar_query(db_cls):
  """Querying for a binding in the dataset."""

  d = read("""a(b). a(c).""", db_cls=db_cls)

  assert select(d, ("a", "X")) == \
    [(("a", "b"), {"X": "b"}),
     (("a", "c"), {"X": "c"})]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_lvar_unification(db_cls):
  """Querying for MATCHING bindings in the dataset."""

  d = read("""edge(b, c). edge(c, c).""", db_cls=db_cls)

  assert select(d, ("edge", "X", "X",)) == \
    [(("edge", "c", "c"), {"X": "c"})]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_rule_join(db_cls):
  """Test a basic join query - the parent -> grandparent relation."""

  child = Constant("child")
  gc = Constant("grandchild")

  d = read("""
child(a, b).
child(b, c).
child(b, d).
child(b, e).

grandchild(A, B) :-
  child(A, C),
  child(C, B).
""", db_cls=db_cls)

  assert select(d, ("grandchild", "a", "X",)) == \
    [(("grandchild", "a", "c"), {"X": "c"}),
     (("grandchild", "a", "d"), {"X": "d"}),
     (("grandchild", "a", "e"), {"X": "e"})]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_antijoin(db_cls):
  """Test a query containing an antijoin."""

  d = read("""
a(foo, bar).
b(foo, bar).
a(baz, qux).
% matching b(baz, qux). is our antijoin test

no-b(X, Y) :-
  a(X, Y),
  ~b(X, Z).
""", db_cls=db_cls)

  assert select(d, ("no-b", "X", "Y")) == \
    [(("no-b", "baz", "qux"),
      {"X": "baz",
       "Y": "qux"})]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_nested_antijoin(db_cls):
  """Test a query which negates a subquery which uses an antijoin.

  Shouldn't exercise anything more than `test_antjoin` does, but it's an interesting case since you
  actually can't capture the same semantics using a single query. Antijoins can't propagate positive
  information (create lvar bindings) so I'm not sure you can express this another way without a
  different evaluation strategy.

  """

  d = read("""
a(foo, bar).
b(foo, bar).
a(baz, qux).
b(baz, quack).

b-not-quack(X, Y) :-
  b(X, Y),
  ~=(Y, quack).

a-no-nonquack(X, Y) :-
  a(X, Y),
  ~b-not-quack(X, Y).
""", db_cls=db_cls)

  assert select(d, ("a-no-nonquack", "X", "Y")) == \
    [(("a-no-nonquack", "baz", "qux"),
      {"X": "baz",
       "Y": "qux"})]


@pytest.mark.parametrize("db_cls,", DBCLS)
def test_alternate_rule(db_cls):
  """Testing that both recursion and alternation work."""

  d = read("""
edge(a, b).
edge(b, c).
edge(c, d).
edge(d, e).
edge(e, f).

path(A, B) :-
  edge(A, B).

path(A, B) :-
  edge(A, C),
  path(C, B).
""", db_cls=db_cls)

  # Should be able to recurse to this one.
  assert select(d, ("path", "a", "f")) == \
    [(("path", "a", "f"), {})]



"""Debris."""

def shuffled(seq):
  """Because random.shuffle() is in-place >.>"""
  s = seq.copy()
  shuffle(s)
  return s


def constexpr_p(expr):
  """Predicate. True of all terms of the expr are constants."""

  return all(isinstance(e, LVar) for e in expr)


class Timing(object):
  """
  A context manager object which records how long the context took.
  """

  def __init__(self):
    self.start = None
    self.end = None

  def __enter__(self):
    from datetime import datetime
    self.start = datetime.utcnow()
    return self

  def __exit__(self, type, value, traceback):
    from datetime import datetime
    self.end = datetime.utcnow()

  def __call__(self):
    """
    If the context is exited, return its duration. Otherwise return the duration "so far".
    """

    from datetime import datetime
    if self.start and self.end:
      return self.end - self.start
    else:
      return datetime.utcnow() - self.start

  def __str__(self):
    return str(self())



from setuptools import setup

setup(
    name="arrdem.datalog",
    # Package metadata
    version="0.0.7",
    license="MIT",
    description="A Datalog engine",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Reid 'arrdem' McKenzie",
    author_email="me@arrdem.com",
    url="https://git.arrdem.com/arrdem/datalog-py",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Database :: Front-Ends",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],

    # Package setup
    package_dir={"": "src"},
    packages=[
        "datalog",
    ],
    scripts=["bin/datalog"],
    install_requires=[
        "yaspin==0.14.3",
    ],
)
